//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


// -------------------------------------------------------------------
// Nes APU.cpp
// Implementation of NES sound unit
// -------------------------------------------------------------------
//	The APU is kind of tricky to emulate and there are a lot of things
// going on in the code, so this is a little overview to get the general
// idea of the emulation.
//
//	The APU has 5 sound channels - 2 squares, a triangle, noise, and a DMC
// for playing sound samples.  All of these channels are based around 
// counters that get clocked (decremented) at a certain frequency, with the
// smallest being 240 Hz (i.e. 4 times a single emulation frame).  At any 
// given time, each channel produces an output, and these outputs are mixed
// together.  The square, triangle, and noise channels range from 0...15, 
// and the DMC ranges from 0...127.  So, the task of this class is to 
// emulate these sound channels, produces their proper output levels, and 
// then mix them together and produce a final sample for our sound buffer.
//
//	The final step of mixing the outputs together was implemented directly
// from a document posted by blaarg on the nesdev boards.  It turns out that
// simply combining the outputs of the respective channels linearly does
// not produce good results, and blaarg has figured out how to scale the
// output level properly using 2 look-up tables to approximate.
// 
//	The final technique used is oversampling.  The APU runs at the same
// speed as the CPU (1.79 MHZ), so technically it generates 1.79 million
// samples per second - way too many to play through a soundcard.  So, 
// suppose we sample at 44 KHZ by taking samples at evenly spaced intervals.
// This is ok, but might not capture higher frequencies.  To try to capture
// these higher frequencies, we can oversample internally, and then downsample
// to 44 KHZ.  So, if the oversample rate is 4, then internally the APU will
// calculate samples at 176 KHZ and then downsample them to 44 KHZ before 
// paying on the audio card.  This seems to help the sound quality.
// -------------------------------------------------------------------

#include "NesAPU.h"






// -------------------------------------------------------------------
// Procedure: Constructor
// -------------------------------------------------------------------
// Description: The constructor sets some default values, builds some look-up tables,
// and initializes the output buffer according to thr sound settings.            
// -------------------------------------------------------------------
NesAPU::NesAPU() {	
	mCpu = NULL;
	mExpChip = NULL;
	
	mintActualSamplesPerFrame = 0;
	
	mboolSoundEnabled = true;		// Sound is enabled
	mintMasterVolume = 3000;		// Volume is pretty loud
	mintSampleRate = 44100;			// Sample rate is 44 KhZ
	mintSampleSize = 16;			// 16-bit sample size
	mintOSRate = 4;					// Oversample rate
	mintClockDivider = 4;	
	mPlayBuffer = NULL;	

	// Set length counter lookup-table
	intLengthTable[0] = 0x05;	intLengthTable[16] = 0x06;
	intLengthTable[2] = 0x0A;	intLengthTable[18] = 0x0C;
	intLengthTable[4] = 0x14;	intLengthTable[20] = 0x18;
	intLengthTable[6] = 0x28;	intLengthTable[22] = 0x30;
	intLengthTable[8] = 0x50;	intLengthTable[24] = 0x60;
	intLengthTable[10] = 0x1E;	intLengthTable[26] = 0x24;
	intLengthTable[12] = 0x07;	intLengthTable[28] = 0x08;
	intLengthTable[14] = 0x0D;	intLengthTable[30] = 0x10;
	intLengthTable[1] = 0x7F;	intLengthTable[17] = 0x09;
	intLengthTable[3] = 0x01;	intLengthTable[19] = 0x10;
	intLengthTable[5] = 0x02;	intLengthTable[21] = 0x0A;
	intLengthTable[7] = 0x03;	intLengthTable[23] = 0x0B;
	intLengthTable[9] = 0x04;	intLengthTable[25] = 0x0C;
	intLengthTable[11] = 0x05;	intLengthTable[27] = 0x0D;
	intLengthTable[13] = 0x07;	intLengthTable[29] = 0x0E;
	intLengthTable[15] = 0x08;	intLengthTable[31] = 0x0F;

	// Set noise channel cycles lookup-table
	intNoiseCycles[0]  = 0x002;	intNoiseCycles[8]   = 0x065;
	intNoiseCycles[1]  = 0x004;	intNoiseCycles[9]   = 0x07F;
	intNoiseCycles[2]  = 0x008;	intNoiseCycles[10]  = 0x0BE;
	intNoiseCycles[3]  = 0x010;	intNoiseCycles[11]  = 0x0FE;
	intNoiseCycles[4]  = 0x020;	intNoiseCycles[12]  = 0x17D;
	intNoiseCycles[5]  = 0x030;	intNoiseCycles[13]  = 0x1FC;
	intNoiseCycles[6]  = 0x040;	intNoiseCycles[14]  = 0x3F9;
	intNoiseCycles[7]  = 0x050;	intNoiseCycles[15]  = 0x7F2;

	// Set DMA frequency lookup-table
	// (NTSC)
	intDMACycles[0][0] = 0xD60; intDMACycles[0][1] = 0xBE0;
	intDMACycles[0][2] = 0xAA0; intDMACycles[0][3] = 0xA00;
	intDMACycles[0][4] = 0x8F0; intDMACycles[0][5] = 0x7F0;
	intDMACycles[0][6] = 0x710; intDMACycles[0][7] = 0x6B0;
	intDMACycles[0][8] = 0x5F0; intDMACycles[0][9] = 0x500;
	intDMACycles[0][10] = 0x470; intDMACycles[0][11] = 0x400;
	intDMACycles[0][12] = 0x350; intDMACycles[0][13] = 0x2A8;
	intDMACycles[0][14] = 0x240; intDMACycles[0][15] = 0x1B0;
	// (PAL)
	intDMACycles[1][0] = 0xC60; intDMACycles[1][1] = 0xB00;
	intDMACycles[1][2] = 0x9D0; intDMACycles[1][3] = 0x940;
	intDMACycles[1][4] = 0x840; intDMACycles[1][5] = 0x750;
	intDMACycles[1][6] = 0x680; intDMACycles[1][7] = 0x630;
	intDMACycles[1][8] = 0x580; intDMACycles[1][9] = 0x4A0;
	intDMACycles[1][10] = 0x410; intDMACycles[1][11] = 0x3B0;
	intDMACycles[1][12] = 0x310; intDMACycles[1][13] = 0x270;
	intDMACycles[1][14] = 0x210; intDMACycles[1][15] = 0x190;


	// Build mixing tables for producing final output
	for (int i = 0; i < 32; i++) {
		square_table[i] = 95.52f / (8128.0f / (float) i + 100.0f);
	}
	for (int j = 0; j < 204; j++) {
		tnd_table[j] = 163.67f / (24329.0f / (float) j + 100.0f);
	}

	intMachineMode = -1;	
	setMachineMode(NTSCNESMODE);
	blnForceExpMute = false;
	//Init();
	//reset();
}




// -------------------------------------------------------------------
// Procedure: Deconstructor
// -------------------------------------------------------------------
// Description:  Delete sound buffers
// -------------------------------------------------------------------
NesAPU::~NesAPU() {
	if (mPlayBuffer != NULL) delete mPlayBuffer;
	mPlayBuffer = NULL;

	if (mExpChip != NULL) delete mExpChip;
	mExpChip = NULL;
}






// -------------------------------------------------------------------
// Procedure: exec
// -------------------------------------------------------------------
// Description:  Executed the specified number of APU cycles.  This routine
// clocks a frame cycle counter, and if enough cycles have elapsed the
// APU frame counter gets clocked.  This routine also clocks a sample
// counter, and if enough cycles have elapsed, a new sample is generated.
// Finally, an oversample counter is also clocked from within this sampling
// portion, and if enough "oversamples" have been computed, then a sample
// is output to the final sound buffer.
// -------------------------------------------------------------------
void NesAPU::exec(int intCycles) {	
	if (!mboolSoundEnabled) return;		// Don't waste time if sound is disabled

	// Run this many APU cycles...
	mintCycleCtr += intCycles; 	
	while (mintCycleCtr > 0) {
		mintCycleCtr--;
	
		// Clock the APU frame cycle counter
		mintFrameCycleCtr--;
		if (!mintFrameCycleCtr) {
			// When the frame cycle counter hits 0, clock the APU frame counter
			clockFrameCounter();
			mintFrameThirdCtr--;

			// Reset the frame cycle counter
			/*if (!mintFrameThirdCtr) {
				// compensate for a frame being a non-integral number of cycles (i.e. 7423.33)
				// 1/3 of time
				//mintFrameCycleCtr = (mintClockDivider == 4) ? 7424 : 5938;
				mintFrameCycleCtr = 7424;
				mintFrameThirdCtr = 3;
			} else {
				// 2/3 of time				
				//mintFrameCycleCtr = (mintClockDivider == 4) ? 7423 : 5939;
				mintFrameCycleCtr = 7423;
			}*/
			mintFrameCycleCtr = intCyclesPerFrame;
		}

		// Clock the sample counter; when enough cycles pass by, grab the output of each channel,
		// normalize according to blaarg's notes, and add to sample accumulator
		mintSampleCtr--;
		if (!mintSampleCtr) {
			// Run each channel for the amount of cycles, so we have up-to-date
			// output levels for each
			execDMC(mintCyclesPerSample);			
			if (sq1.blnEnabled) execSquare(sq1, mintCyclesPerSample);
			if (sq2.blnEnabled) execSquare(sq2, mintCyclesPerSample);
			if (tri.blnEnabled) execTriangle(mintCyclesPerSample);
			if (nse.blnEnabled) execNoise(mintCyclesPerSample);
									
			// Run expansion chip
			if (mExpChip) mExpChip->exec(mintCyclesPerSample);

			// Mute channels according to client's whim
			if (sq1.blnForceMute) sq1.intCurSample = 0;
			if (sq2.blnForceMute) sq2.intCurSample = 0;			
			if (tri.blnForceMute) tri.intCurSample = 0;
			if (nse.blnForceMute) nse.intCurSample = 0;
			if (dmc.blnForceMute) dmc.intCurSample = 0;
			//if (mExpChip && blnForceExpMute) mExpChip->setCurrentSample(0);


			// Normalize the output to 0...1 and add to oversample accumulator
			fltSample += square_table[sq1.intCurSample + sq2.intCurSample] + tnd_table[ 3*tri.intCurSample + 2*nse.intCurSample + dmc.intCurSample];

			// Clock the oversample counter; when enough oversamples have accumulated
			// take the average of the oversample and store it in the output buffer
			mintOSCtr--;
			if (!mintOSCtr) {				
				float avgSample = fltSample / mintOSRate;				
				int sample = avgSample * mintMasterVolume * 10;
					
				// Add expansion chip output into mix
				if (mExpChip) {
					// Sample is in range 0...64 (6 bits)
					int newSample = blnForceExpMute ? 0 : mExpChip->getCurrentSample();										
					int add = (newSample * mintMasterVolume * 10) / 64;
					sample = (sample + add);
					if (sample > mintMasterVolume * 10) sample = mintMasterVolume * 10;
				} 
				//sample = sample << 1;
				//if (sample > mintMasterVolume * 10) sample = mintMasterVolume * 10;
				

				if (mintOutputCtr < mintActualSamplesPerFrame) {
					// Write out appropriately-sized sample
					switch (mintSampleSize) {
					case 8:
						{
							Uint8 *outputPtr = (Uint8 *)mPlayBuffer;
							outputPtr[ mintOutputCtr ] = sample >> 8;
						}
						break;
					case 16:
						{
							Uint16 *outputPtr = (Uint16 *)mPlayBuffer;
							outputPtr[ mintOutputCtr ] = sample;
						}
						break;
					}
				
					mintOutputCtr++;
				}
				
				mintOSCtr = mintOSRate;
				fltSample = 0.0f;
				
			}
		
			// Reset the sample counter
			mintSampleCtr = mintCyclesPerSample;
		}
	}
}






// -------------------------------------------------------------------
// Procedure: execSquare
// -------------------------------------------------------------------
// Description: This function executes the square channel for the given
// number of cycles.  It computes the output level of the square
// channel.  The output level toggles between 0 and 1 (* the volume) at
// a frequency determined by the wavelength, and this function basically
// makes sure that the output level follows this frequency
// -------------------------------------------------------------------
void NesAPU::execSquare(SquareChannel &sq, int intCycles) {	
	if ((sq.intLengthCtr) && (sq.intWavelength >= 0x08)) {
		int intCycleCtr = 0;		
		while (intCycleCtr < intCycles) {
			// Clock the square cycle counter, upon terminal count do something
			sq.intCycleCtr--;
			if (!sq.intCycleCtr) {
				// Update the sample 
				int DutyLimit = 2;
				if (sq.intDutyType > 0) DutyLimit = sq.intDutyType * 4;
				
				if (sq.intDutyCtr < DutyLimit) {
					sq.intCurSample = sq.intCurVol;
				} else {					
					sq.intCurSample = 0;
				}
				sq.intDutyCtr = (sq.intDutyCtr + 1) & 0x0F;

				// Reset cycle counter
				sq.intCycleCtr = sq.intWavelength+1;
			}
			intCycleCtr++;
		}		
	}
}

// -------------------------------------------------------------------
// Procedure: execTriangle
// -------------------------------------------------------------------
// Description:  The function executes the triangle channel for a given
// number of clock cycles, computing its output.  The triangle channel
// follows a triangle wave pattern, and shifts to the next phase in the
// triangle pattern every "wavelength" number of cycles, which is what 
// this routine is in charge of doing.
// -------------------------------------------------------------------
void NesAPU::execTriangle(int intCycles) {		
	if (tri.intCycleCtr && (tri.intLinearCtr && tri.blnEnabled)) {
		if (tri.intWavelength > 0x08) {			
			int intCycleCtr = 0;
			while (intCycleCtr < intCycles) {
				tri.intCycleCtr--;
				if (!tri.intCycleCtr) {
					// Set output
					tri.intCurSample = tri.intStepCtr;

					// Generate triangle step
					if (tri.intStepCtr == TRIMIN) {							
						tri.intStepInc++;								
					} else if (tri.intStepCtr == TRIMAX) {
						tri.intStepInc--;	
					}											
					tri.intStepCtr += tri.intStepInc;
										
					// Reset counter
					tri.intCycleCtr = tri.intWavelength + 1;
				}				
				intCycleCtr++;
			}
		} 				
	} else {
		tri.intCurSample = 0;
	}
}






// -------------------------------------------------------------------
// Procedure: execNoise
// -------------------------------------------------------------------
// Description:  This function executes the noise channel for a given
// number of clock cycles.  The noise channel basically produces
// waves with random durations, producing static-y sound.  The core
// of the noise channel is a random number generator described in
// detail in Brad Taylor's NESSOUND document.
// -------------------------------------------------------------------
void NesAPU::execNoise(int intCycles) {
	//if (nse.intLengthCtr && (nse.intWavelength > 0x08) ) {
	if (nse.intLengthCtr && nse.intWavelength) {	
	//if (nse.intWavelength) {	
		int intCycleCtr = 0;
		while (intCycleCtr < intCycles) {
			nse.intCycleCtr--;
			if (!nse.intCycleCtr) {
				
				// Random number generator
				int intNewBit0 = 0x0;
				if (nse.blnDataType) {
					// 93 bit mode - new bit 0 is XOR of bits 2 and 8
					intNewBit0 = ((nse.intShiftReg & 0x0004) >> 2) ^ ((nse.intShiftReg & 0x0100) >> 8);
				} else {
					// 32k mode - new bit 0 is XOR of bits E and D
					intNewBit0 = ((nse.intShiftReg & 0x4000) >> 14) ^ ((nse.intShiftReg & 0x2000) >> 13);
				}
				
				nse.intShiftReg = (nse.intShiftReg << 1) | intNewBit0;

				// Invert the last pin of the 15-bit shfit register to get output
				nse.intCurRand = !((nse.intShiftReg & 0x4000) >> 14);
				
				//nse.intCycleCtr = (nse.intWavelength << 1) + 1;
				nse.intCycleCtr = (nse.intWavelength << 1);
			}
			intCycleCtr++;
		}
		nse.intCurSample = (nse.intCurRand * nse.intCurVol);
	} else {
		nse.intCurSample = 0;
	}
}






// -------------------------------------------------------------------
// Procedure: execDMC
// -------------------------------------------------------------------
// Description: This function executes the DMC channel for a given number
// of cycles.  The DMC channel periodically reads from the CPU memory, and 
// based on  the contents of the memory adds or subtracts a a counter that
// basically represents the amplitude of the channel.  So, that's a very
// high-level description, check out Brad Taylor's DMC doc for the real story.
// -------------------------------------------------------------------
void NesAPU::execDMC(int intCycles) {
	if (dmc.blnEnabled) {
		// Countdown the DMA frequency counter if channel is enabled
		int intCycleCtr = 0;	
		while (intCycleCtr < intCycles) {
			if (dmc.intDMALength) {
				dmc.intCycleCtr--;
				if (!dmc.intCycleCtr) {
					// Fetch a new byte from the DMA address
					dmc.intCurByte = mCpu->RAM_Read(dmc.intDMAAddr);
					mCpu->mlngCycleCounter++;
					dmc.intDMAAddr++;
					if (dmc.intDMAAddr == 0xFFFF+1) dmc.intDMAAddr = 0x8000;
					dmc.intDMALength--;
					
					// See if end of sample is hit, react according to playmode
					if (!dmc.intDMALength) {											
						// This reduces clicks, but may not be valid
						//dmc.intDeltaCounter = 0;

						// Look at Playmode
						switch (dmc.intPlayMode) {
						case 0x00: // Play until stopped, halt						
							dmc.blnEnabled = false;							
							break;
						case 0x01: // Loop
							dmc.intDMAAddr = dmc.intDMAAddrLatch;
							dmc.intDMALength = dmc.intDMALengthLatch;						
							break;
						case 0x02:	// Gen IRQ
							{
								// TODO
							}
							break;
						}
					}
					
					// Reset cycle, shift counters
					dmc.intShiftCtr = dmc.intDMARate >> 3;
					dmc.intCycleCtr = dmc.intDMARate;
				} else {
					// If DMA doesn't need to get a byte yet, clock the shift register (at 8X the speed)
					// and update the delta counter upon terminal count
					dmc.intShiftCtr--;
					if (!dmc.intShiftCtr) {
						// Look at LSB and add/subtract delta counter
						if (dmc.intCurByte & 0x01) {
							// Add one
							if (dmc.intDeltaCounter < 0x3F) dmc.intDeltaCounter++;
						} else {
							// Subtract one
							if (dmc.intDeltaCounter > 0) dmc.intDeltaCounter--;
						}										
						dmc.intCurByte >>= 1;
						dmc.intShiftCtr = dmc.intDMARate >> 3;
					}
				}
			}		
			intCycleCtr++;
		}
	}

	// After executing, set the output level regardless if DMC channel is enabled (via $4015)
	// or not (DMC channel can be disabled but still produce output, see the "crowd noise"
	// at beginning of Punch-Out!!	
	dmc.intCurSample = (dmc.intDeltaCounter << 1) | (dmc.intDAC & 0x01);
}





// -------------------------------------------------------------------
// Procedure: clockFrameCounter
// -------------------------------------------------------------------
// Description: Clock the frame counter.  When the frame counter get 
// clocked, various components of the APU receive a clock depending 
// on what the frame counter is
// ------------------------------------------------------------------
void NesAPU::clockFrameCounter() {
	// Depending on what stage of the frame counter...
	switch(mintFrameCounter) {
	case 0:			
		// Clock linear (triangle) and envelope decay (square + noise) counters			
		clockLinearCounter();
		clockVolumeDecay(sq1);
		clockVolumeDecay(sq2);
		clockNoiseVolume();
		break;
		
	case 1:
		// Clock linear (triangle) and envelope decay (square + noise) counters
		clockLinearCounter();
		clockVolumeDecay(sq1);
		clockVolumeDecay(sq2);
		clockNoiseVolume();

		// Clock frequency sweep (square) and length counters
		clockLengthCtr(1);
		clockFreqSweep(sq1, 1);
		
		clockLengthCtr(2);
		clockFreqSweep(sq2, 2);

		clockLengthCtr(3);

		clockLengthCtr(4);
		break;
		
	case 2:
		// Clock linear (triangle) and envelope decay (square + noise) counters			
		clockLinearCounter();
		clockVolumeDecay(sq1);
		clockVolumeDecay(sq2);
		clockNoiseVolume();
		break;
		
	case 3:
		// Clock linear (triangle) and envelope decay (square + noise) counters
		clockLinearCounter();
		clockVolumeDecay(sq1);
		clockVolumeDecay(sq2);
		clockNoiseVolume();

		// Clock frequency sweep (square)			
		clockFreqSweep(sq1, 1);			
		clockFreqSweep(sq2, 2);
		break;

	case 4:
		// Don't clock anything on this cycle.  It only gets hit when clock divide is 5
		break;
	}

	mintFrameCounter++;
	
	if (mintFrameCounter == (mintClockDivider)) {
		// Reset counter
		mintFrameCounter = 0;
			
		// Raise IRQ line if needed TODO: this might be in the wrong place
		if (blnIRQEnabled) {
			//mCpu->mintInterruptLine |= APUIRQ;
		}
	}
}









// -------------------------------------------------------------------
// Procedure: clockLengthCtr
// -------------------------------------------------------------------
// Description: This clocks the length counter for each channel.  The
// length counter dictates how long a channel is "active" - that is, 
// how long the channel's output will be enabled
// -------------------------------------------------------------------
void NesAPU::clockLengthCtr(int channel) {
	switch (channel) {
	case 1: // Square #1
		if (sq1.blnEnabled & sq1.blnLengthEnabled) {			
			if (sq1.intLengthCtr) {
				sq1.intLengthCtr--;	
			}
			if (!sq1.intLengthCtr) {
				sq1.intCurSample = 0;
				sq1.intDutyCtr = 0;
			}
		}		
		break;
	case 2: // Square #2
		if (sq2.blnEnabled & sq2.blnLengthEnabled) {			
			if (sq2.intLengthCtr > 0) {
				sq2.intLengthCtr--;	
			}
			if (!sq2.intLengthCtr) {
				sq2.intCurSample = 0;
				sq2.intDutyCtr = 0;
			}
		}
		break;
	case 3: // Triangle wave
		if (tri.blnEnabled & tri.blnLinearEnable) {
			// Change linear counter mode on linear clock if we have to
			if ((tri.intLinearMode == LOADMODE) && (tri.blnChangeMode)) {
				tri.intLinearMode = COUNTMODE;
				tri.blnChangeMode = false;
			}
			if (tri.intLengthCtr) {
				tri.intLengthCtr--;			
				if (!tri.intLengthCtr) {
					tri.intStepCtr = TRIMIN;
					tri.intStepInc = 0;
					tri.intCurSample = 0;
				}
			}
		}

		break;
	case 4: // Noise channel
		if (nse.blnEnabled & nse.blnLengthEnabled) {
			if (nse.intLengthCtr) nse.intLengthCtr--;
			if (!nse.intLengthCtr) nse.intCurSample = 0;
			nse.intCurSample = 0;
		}		
		break;
	}
}





// -------------------------------------------------------------------
// Procedure: clockVolumeDecay
// -------------------------------------------------------------------
// Description: This clocks the volume/decay component of the square
// channels.
// -------------------------------------------------------------------
void NesAPU::clockVolumeDecay(SquareChannel &sq) {
		/*if (sq.blnDecayEnabled && sq.intLengthCtr) {
			// Clock decay counter
			sq.intDecayRate--;

			// Decay cycle complete, decrement volume by 1
			if (sq.intDecayRate == 0) {
				sq.intDecayRate = sq.intVolume + 1;
				sq.intCurVol = sq.intDecayCtr;
				
				// Look at bit 5 and either wrap around to $F or stay at 0
				if (!sq.intDecayCtr) {
					if (!sq.blnDecayEnabled) {
						sq.intDecayCtr = 0x0F;
					} else {
						sq.intDecayCtr = 0x00;
					}
				} else {
					sq.intDecayCtr--;
				}
			}
		} else {
			//sq.intCurVol = sq.intVolume;
		}*/
	if (sq.blnEnabled && sq.intLengthCtr) {
		// Clock decay counter
		sq.intDecayRate--;

		// Decay cycle complete, decrement volume by 1
		if (!sq.intDecayRate) {
			sq.intDecayRate = sq.intVolume + 1;
							
			// Look at bit 5 and either wrap around to $F or stay at 0
			if (!sq.intDecayCtr) {
				if (!sq.blnLengthEnabled) {
					sq.intDecayCtr = 0x0F;
				} else {
					sq.intDecayCtr = 0x00;
				}
			} else {
				sq.intDecayCtr--;
			}
			sq.intCurVol = (!sq.blnDecayEnabled) ? sq.intVolume : sq.intDecayCtr;
		}
	} else {
		sq.intCurVol = sq.intVolume;
	}

}






// -------------------------------------------------------------------
// Procedure:  clockFreqSweep
// -------------------------------------------------------------------
// Description:  This clocks the sweep component of the square channels
// -------------------------------------------------------------------
void NesAPU::clockFreqSweep(SquareChannel &sq, int channel) {				
	//if (sq.blnEnabled) {	
	if (sq.blnEnabled && sq.intLengthCtr) {	
		if (sq.blnSweepEnabled && ((sq.intWavelength >= 0x08) && (sq.intWavelength <= 0x7FF))) {
			// Clock counter
			sq.intSweepCtr--;
			
			// Cycle has completed, adjust waveform if 3 conditions are met
			if (!sq.intSweepCtr) {
				// Extract the shift amount, see if it's zero					
				if (sq.intShiftAmt) {
					// Update the 11-bit wavelength												
					// Look at bit 3, it tells the direction, and update wavelength
					int sweepAmt = sq.intWavelength >> sq.intShiftAmt;
					int intNewWavelength;
					if (sq.blnSweepDir) {
						// Subtraction; some small variation between square channels
						if (channel == 1) { 
							intNewWavelength = sq.intWavelength + ~(sweepAmt);
						} else {
							intNewWavelength = sq.intWavelength + -(sweepAmt);
						}							
					} else {
						intNewWavelength = sq.intWavelength + (sweepAmt);
					}
					sq.intWavelength = intNewWavelength;					
				}	
				
				// Reset the counter
				sq.intSweepCtr = sq.intSweepRate+1;
			}
		}
	}
}





// -------------------------------------------------------------------
// Procedure:  clockNoiseVolume
// -------------------------------------------------------------------
// Description:  Clocks the volume/decay of the noise channel
// -------------------------------------------------------------------
void NesAPU::clockNoiseVolume() {
	//if (nse.blnDecayEnabled) {
	if (nse.blnEnabled && nse.intLengthCtr) {
		// Clock decay counter
		nse.intDecayRate--;

		// Decay cycle complete, decrement volume by 1
		if (!nse.intDecayRate) {
			nse.intDecayRate = nse.intVolume + 1;
							
			// Look at bit 5 and either wrap around to $F or stay at 0
			if (!nse.intDecayCtr) {
				if (!nse.blnLengthEnabled) {
					nse.intDecayCtr = 0x0F;
				} else {
					nse.intDecayCtr = 0x00;
				}
			} else {
				nse.intDecayCtr--;
			}
			nse.intCurVol = (!nse.blnDecayEnabled) ? nse.intVolume : nse.intDecayCtr;
		}
	} else {
		nse.intCurVol = nse.intVolume;
	}
}





// -------------------------------------------------------------------
// Procedure: clockLinearCounter
// -------------------------------------------------------------------
// Description:  Clocks the linear counter component of the triangle
// channel. 
// -------------------------------------------------------------------
void NesAPU::clockLinearCounter() {
	// If we are in count mode, then count!
	if (tri.intLinearMode == COUNTMODE) {
		if (tri.intLinearCtr) {
			tri.intLinearCtr--;
			if (!tri.intLinearCtr) {
				tri.intStepCtr = TRIMIN;
				tri.intStepInc = 0;
				tri.intCurSample = 0;
			}
		}
	}
}




// -------------------------------------------------------------------
// Procedure: read 
// -------------------------------------------------------------------
// Description: This is the CPU read function.  The CPU can read the APU
// to query the status of the various APU channels
// -------------------------------------------------------------------
// Memory read function
unsigned char NesAPU::read(Uint16 addr) {	
	switch (addr) {
	case 0x4015:		
		{
			Uint8 retval= 0x00;
			if (sq1.intLengthCtr != 0) retval |= 0x01;	// Sq #1 status
			if (sq2.intLengthCtr != 0) retval |= 0x02;	// Sq #2 status
			if (tri.intLengthCtr != 0) retval |= 0x04;	// Tri status
			if (nse.intLengthCtr != 0) retval |= 0x08;	// Noise status
			if (dmc.blnEnabled) retval |= 0x10;			// DMC playing sample?
						
			return retval;
		}
		break;
	}
	return 0;
}





// -------------------------------------------------------------------
// Procedure: write
// -------------------------------------------------------------------
// Description: This is the interface routine for the CPU/APU.  The CPU
// writes to these registers to set various properties of the APU
// -------------------------------------------------------------------
void NesAPU::write(Uint16 addr, unsigned char data) {	
	switch (addr) {
	/////////// Square Channel #1 /////////////////
	case 0x4000:	// Square channel #1 volume/decay, enabling bits				
		sq1.intDutyType = (data & 0xC0) >> 6;				// Save duty cycle type		
		sq1.blnLengthEnabled = (bool) !((data & 0x20) >> 5);
		sq1.blnDecayEnabled = (bool) !((data & 0x10) >> 4);		
		sq1.intVolume = data & 0x0F;

		if (sq1.blnDecayEnabled) {
			sq1.intCurVol = sq1.intDecayCtr;
		} else {
			sq1.intCurVol = sq1.intVolume;
		}

		sq1.intDecayRate = sq1.intVolume + 1;
		break;

	case 0x4001:	// Square channel #1 sweep/shift register				
		// Bit 7 of data is enabled bit for sweeping
		sq1.blnSweepEnabled = (bool) ((data & 0x80) >> 7);
		sq1.intSweepRate = ((data & 0x70) >> 4);
		sq1.intShiftAmt = data & 0x7;						// Save the shift amount
		sq1.blnSweepDir = (bool) ((data & 0x08) >> 3);		// Save sweep direction
		break;

	case 0x4002:	// Square channel #1 wavelength			
		sq1.intWavelength = (sq1.intWavelength & 0x700) | data;
		break;

	case 0x4003:	// Square channel #1 length counter		
		if (sq1.blnEnabled) {
			sq1.intLengthCtr = intLengthTable[ (data & 0xF8)>>3 ];		
		}
		sq1.intWavelength = (sq1.intWavelength & 0xFF) | ((data & 0x7) << 8);
		sq1.intDecayCtr = 0x0F;								// Reset envelope decay counter		
		break;



	/////////// Square Channel #2 ////////////////////
	case 0x4004:		
		sq2.intDutyType = (data & 0xC0) >> 6;				// Save duty cycle type		
		sq2.blnLengthEnabled = (bool) !((data & 0x20) >> 5);
		sq2.blnDecayEnabled = (bool) !((data & 0x10) >> 4);		
		sq2.intVolume = data & 0x0F;

		if (sq2.blnDecayEnabled) {
			sq2.intCurVol = sq2.intDecayCtr;
		} else {
			sq2.intCurVol = sq2.intVolume;
		}		
		break;

	case 0x4005:		
		// Bit 7 of data is enabled bit for sweeping
		sq2.blnSweepEnabled = (bool) ((data & 0x80) >> 7);
		sq2.intSweepRate = ((data & 0x70) >> 4);
		sq2.intShiftAmt = data & 0x7;						// Save the shift amount
		sq2.blnSweepDir = (bool) ((data & 0x08) >> 3);		// Save sweep direction		
		break;

	case 0x4006:				
		sq2.intWavelength = (sq2.intWavelength & 0x700) | data;		
		break;

	case 0x4007:				
		sq2.intLengthCtr = intLengthTable[ (data & 0xF8)>>3 ];		
		sq2.intWavelength = (sq2.intWavelength & 0xFF) | ((data & 0x7) << 8);		
		sq2.intDecayCtr = 0x0F;								// Reset envelope decay counter		
		break;


	//////////////// Triangle Channel Registers ///////////////////////
	case 0x4008:	// Triangle channel linear counter register				
		tri.blnLinearEnable = (bool) !((data & 0x80) >> 7);
		//if (!tri.blnLengthEnabled) tri.intCurSample = 0;
		tri.intLinearLimit = (data & 0x7F);
		if (tri.intLinearMode == LOADMODE) {
			// TODO: iffy
			if (tri.intLengthCtr) 
				tri.intLinearCtr = tri.intLinearLimit;				
		}
		break;

	case 0x400A:	// Triangle channel wavelength register	
		tri.intWavelength = (tri.intWavelength & 0x700) | data;		
		break;

	case 0x400B:	// Triangle channel MSB wavelength, length counter		
		tri.intWavelength = (tri.intWavelength & 0xFF) | ((data & 0x7) << 8);				
		tri.intLengthCtr = intLengthTable[ (data & 0xF8)>>3 ];
		tri.intLinearMode = LOADMODE;
		tri.intLinearCtr = tri.intLinearLimit;
		tri.blnChangeMode = true;
		break;


	///////////////// Noise Channel Registers ////////////////	
	case 0x400C:	// Noise channel volume/decay register
		nse.blnDecayEnabled = (bool) !((data & 0x10) >> 4);
		nse.blnLengthEnabled = (bool) !((data & 0x20) >> 5);
		nse.intVolume = data & 0x0F;
		if (nse.blnDecayEnabled) {
			//nse.intCurVol = nse.intDecayRate;
			nse.intCurVol = nse.intDecayCtr;
		} else {
			nse.intCurVol = nse.intVolume;
		}		
		break;
		
	case 0x400E:	// Noise channel lookup value register		
		nse.intWavelength = intNoiseCycles[ data & 0x0F ];		
		nse.blnDataType = (bool) ((data & 0x80) >> 7);		
		break;
	
	case 0x400F:	// Noise channel length counter register
		if (nse.blnEnabled) { 
			nse.intLengthCtr = intLengthTable[ (data & 0xF8) >> 3 ];		
			nse.intLengthLimit = nse.intLengthCtr;
		}
		nse.intDecayCtr = 0x0F;		
		break;

	//////////////////// Delta Modulation Channel (DMC) registers ////////////////
	
	// Play mode and DMA frequency register
	case 0x4010:
		dmc.intPlayMode = (data & 0xC0) >> 6;
		dmc.intDMARate = intDMACycles[intMachineMode][ (data & 0xF) ];		
		dmc.intShiftCtr = dmc.intDMARate >> 3;
		break;
	
	// Delta counter load register
	case 0x4011:
		dmc.intDAC = (dmc.intDAC & 0x7E) | (data & 0x01);
		dmc.intDeltaCounter = (data & 0x7E) >> 1;
		break;
	
	// DMA address load register
	case 0x4012:
		dmc.intDMAAddr = dmc.intDMAAddrLatch = (data << 6) | 0xC000;
		break;
	
	// DMA length register
	case 0x4013:
		dmc.intDMALength = dmc.intDMALengthLatch = data << 4;
		break;


	case 0x4015:	// Channel enable register			
		sq1.blnEnabled = (bool) (data & 0x01);
		sq2.blnEnabled = (bool) ((data & 0x02) >> 1);
		tri.blnEnabled = (bool) ((data & 0x04) >> 2);
		nse.blnEnabled = (bool) ((data & 0x08) >> 3);
				
		if (!sq1.blnEnabled) {
			sq1.intLengthCtr = 0;
			sq1.intCurSample = 0;			
		} 
		if (!sq2.blnEnabled) {			
			sq2.intLengthCtr = 0;
			sq2.intCurSample = 0;			
		}
		if (!tri.blnEnabled) {			
			tri.intLinearCtr = 0;
			tri.intLengthCtr = 0;
			tri.intCurSample = 0;
			tri.intStepCtr = TRIMIN;
			tri.intStepInc = 0;	
		}
		if (!nse.blnEnabled) {			
			nse.intLengthCtr = 0;
			nse.intCurSample = 0;			
		}
		
		if ((data & 0x10) >> 4) {
			if (!dmc.blnEnabled) {
				dmc.intDMAAddr = dmc.intDMAAddrLatch;
				dmc.intDMALength = dmc.intDMALengthLatch;
				dmc.intCurSample = dmc.intDeltaCounter = 0;				
			}
			dmc.blnEnabled = true;
		} else {
			dmc.intCurSample = 0;
			dmc.blnEnabled = false;
		}

		// Take down frame counter IRQ flag
		mCpu->mintInterruptLine &= (!APUIRQ);		
		break;

	case 0x4017:		
		// Reset frame counter, clock divider
		mintFrameCounter = 0;		
		if (data & 0x80) {
			mintClockDivider = 5;
			mintFrameCycleCtr = 1;
			//mintFrameCounter = 4;	// Start with cycle 4
		} else {			
			mintClockDivider = 4;
			//mintFrameCycleCtr = 7424;
			mintFrameCycleCtr = intCyclesPerFrame;
		}
		
		//blnIRQEnabled = ! ((data & 0x40) >> 5);				
		//clockFrameCounter();
		break;
	
	default:
		break;

	}	
}






// -------------------------------------------------------------------
// Procedure: endFrame
// -------------------------------------------------------------------
// Description: This makes sure that the output buffer gets filled with
// enough samples to fill out a frame (just in case!)
// -------------------------------------------------------------------
// Runs the APU until the end of the frame
void NesAPU::endFrame() {	
	if (!mboolSoundEnabled) return;		// Don't waste time if sound is disabled
	
	/*while (mintOutputCtr != 0) {		
		exec(1);
	}*/
	mintOutputCtr = 0;
}






// -------------------------------------------------------------------
// Procedure: initSquareChannel
// -------------------------------------------------------------------
// Description: Initializes a square channel
// -------------------------------------------------------------------
void NesAPU::initSquareChannel(SquareChannel &sq) {
	sq.intLengthCtr = 0;
	sq.intDecayCtr = 1;
	sq.intDecayRate = 1;
	sq.intDutyCtr = 0;
	sq.intVolume = 0xF;
	sq.intWavelength = 0;
	sq.intShiftAmt = 0;
	sq.blnForceMute = false;
	sq.intCurSample = 0;
	sq.intCurVol = 0;
}




// -------------------------------------------------------------------
// Procedure: Init
// -------------------------------------------------------------------
// Description: Does some initalization of the APU
// -------------------------------------------------------------------
void NesAPU::Init() {
	initSquareChannel(sq1);
	initSquareChannel(sq2);
		
	tri.blnEnabled = false;	
	tri.intStepCtr = TRIMIN;
	tri.intStepInc = 0;	
	tri.blnForceMute = false;	
	tri.intCurSample = 0;
		
	memset(&nse, 0x00, sizeof(nse));
	nse.intCurRand = 1;		
	
	dmc.blnForceMute = dmc.blnEnabled = false;	
	dmc.intCurSample = 0;
	dmc.intDMAAddr = dmc.intDMAAddrLatch = dmc.intDMALength = dmc.intDMALengthLatch = 0;
	dmc.intCurSample = 0;

	initSoundBuffers();
	
	mintFrameCounter = 0;
	mintClockDivider = 4;
}



// -------------------------------------------------------------------
// Procedure: reset
// -------------------------------------------------------------------
// Description: This resets the APU, including all of its channels
// -------------------------------------------------------------------
void NesAPU::reset() {
	sq1.intCurSample = sq2.intCurSample = nse.intCurSample = tri.intCurSample = dmc.intCurSample = 0;
	sq1.intCycleCtr = sq2.intCycleCtr = tri.intCycleCtr = nse.intCycleCtr = dmc.intCycleCtr = 1;
	sq1.blnEnabled = sq2.blnEnabled = tri.blnEnabled = nse.blnEnabled = dmc.blnEnabled = false;

	sq1.intSweepCtr = sq2.intSweepCtr = 1;
	sq1.intDecayRate = sq2.intDecayRate = 1;
	sq1.blnSweepEnabled = sq2.blnSweepEnabled = false;
	sq1.intShiftAmt = sq2.intShiftAmt = 0;
	blnIRQEnabled = true;
	
	// Reset DMC channel
	dmc.blnEnabled = false;
	dmc.intDAC = dmc.intDeltaCounter = 0;

	tri.intStepCtr = TRIMIN;
	tri.intStepInc = 0;	
	tri.intLinearMode = COUNTMODE;
	tri.intLinearCtr = tri.intLinearLimit = 0;
	tri.blnChangeMode = false;

	nse.intDecayCtr = 0;
	nse.intShiftReg = 0x0001;	
	nse.intDecayRate = 1;	
	
	mintFrameThirdCtr = 3;
	mintFrameCycleCtr = 1;
	mintSampleCtr = 1;
	mintOSCtr = 1;
	mintOutputCtr = 0;
	fltSample = 0.0f;
	
	//memset(mPlayBuffer, 0x00, mintBufferSize * (mintSampleSize >> 3) );
	memset(mPlayBuffer, 0x00, mintActualSamplesPerFrame * (mintSampleSize >> 3) );

	if (mExpChip) mExpChip->reset();
}





// -------------------------------------------------------------------
// Procedure: linkCPU
// -------------------------------------------------------------------
// Description: This links the APU to the CPU.  The APU needs access to
// the CPU for DMC (so the DMC channel can read CPU memory), and to fire
// IRQs
// -------------------------------------------------------------------
void NesAPU::linkCPU(NesCPU *linkCPU) {
	mCpu = linkCPU;
}






// -------------------------------------------------------------------
// Procedure: initSoundBuffers
// -------------------------------------------------------------------
// Description: Deletes sound buffers and re-allocates them, typically 
// called internally after a samplerate or samplesize change
// -------------------------------------------------------------------
void NesAPU::initSoundBuffers() {
	if (mPlayBuffer != NULL) delete mPlayBuffer;	

	// Divide by 59 rather than 60; this will allow for any round-off error
	// and allocate enough storage for the desired number of buffered frames
	//mintBufferSize = (mintSampleRate / (intWantFPS-1)); 	

	//mPlayBuffer = new char[mintBufferSize * (mintSampleSize >> 3)];	
	//memset(mPlayBuffer, 0x00, mintBufferSize * (mintSampleSize >> 3) );
	
	mintCyclesPerSample = (intCPUSpeed / (mintSampleRate * mintOSRate) + .5);	
	
	// This is important:  APU is guaranteed to get 29780 clocks/frame
	// So we make sure to only play 29780/cyclesPerSample on the soundcard;
	// it is important that these match as close as possible!
	if (mCpu != NULL) {						
		mintActualSamplesPerFrame = (mCpu->MTimers.intCyclesPerFrame) / (mintCyclesPerSample *  mintOSRate);		
	} else {
		mintActualSamplesPerFrame = 1;
	}
	
	mPlayBuffer = new char[mintActualSamplesPerFrame * (mintSampleSize >> 3)];	
	memset(mPlayBuffer, 0x00, mintActualSamplesPerFrame * (mintSampleSize >> 3) );

	mintSampleCtr = 1;
	mintCycleCtr = 0;		
}




// -------------------------------------------------------------------
// Procedure: setMachineMode
// -------------------------------------------------------------------
// Description: Sets the timing settings for NTSC/PAL emulation
// -------------------------------------------------------------------
void NesAPU::setMachineMode(int intNewMode) {
	if (intMachineMode != intNewMode) {
		switch (intNewMode) {
		case NTSCNESMODE:
			intCPUSpeed = 1789772;
			intCyclesPerFrame = 7424;
			intWantFPS = 60;
			break;
		case PALNESMODE:
			intCPUSpeed = 1662607;
			intCyclesPerFrame = 8313;
			intWantFPS = 50;
			break;
		}

		Init();
		reset();
		intMachineMode = intNewMode;
	}
}


// -------------------------------------------------------------------
// Procedure: addExpansionChip
// -------------------------------------------------------------------
// Description: Adds an expansion sound chip to the APU.  The expansion
// chip will then be emulated and mixed into the final APU output.
// -------------------------------------------------------------------
void NesAPU::addExpansionChip(int intExpChip) {
	if (mExpChip) delete mExpChip;
	switch (intExpChip) {
	case VRC6CHIP:	// Konami VRC6 - Used in Castlevania 3 (J), Madara
		mExpChip = new VRC6();
		break;
	}
}


// -------------------------------------------------------------------
// Procedure: writeExpansionChip
// -------------------------------------------------------------------
// Description: Passes a write along to the expansion chip
// -------------------------------------------------------------------
void NesAPU::writeExpansionChip(Uint16 addr, unsigned char data) {
	if (mExpChip) {
		mExpChip->write(addr, data);
	}
}




//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the APU to the file stream
//-------------------------------------------------------------------
void NesAPU::saveState(FILE *stream) {
	fwrite( &sq1, sizeof(SquareChannel), 1, stream);	// Square Channel #1
	fwrite( &sq2, sizeof(SquareChannel), 1, stream);	// Square Channel #2
	fwrite( &tri, sizeof(TriangleChannel), 1, stream);	// Triangle Channel
	fwrite( &nse, sizeof(NoiseChannel), 1, stream);		// Noise Channel
	fwrite( &dmc, sizeof(DMCChannel), 1, stream);		// DMC Channel
	fwrite( &mintFrameCounter, sizeof(mintFrameCounter), 1, stream);    // APU Frame counter
	fwrite( &mintFrameCycleCtr, sizeof(mintFrameCycleCtr), 1, stream);  // APU Frame cycle counter
	fwrite( &mintFrameThirdCtr, sizeof(mintFrameThirdCtr), 1, stream);  // APU Frame cycle counter counter
	
	if (mExpChip) mExpChip->saveState(stream);
}





//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the APU from the file stream
//-------------------------------------------------------------------
void NesAPU::loadState(FILE *stream) {
	fread( &sq1, sizeof(SquareChannel), 1, stream);		// Square Channel #1
	fread( &sq2, sizeof(SquareChannel), 1, stream);		// Square Channel #2
	fread( &tri, sizeof(TriangleChannel), 1, stream);	// Triangle Channel
	fread( &nse, sizeof(NoiseChannel), 1, stream);		// Noise Channel
	fread( &dmc, sizeof(DMCChannel), 1, stream);		// DMC Channel
	fread( &mintFrameCounter, sizeof(mintFrameCounter), 1, stream);    // APU Frame counter
	fread( &mintFrameCycleCtr, sizeof(mintFrameCycleCtr), 1, stream);  // APU Frame cycle counter
	fread( &mintFrameThirdCtr, sizeof(mintFrameThirdCtr), 1, stream);  // APU Frame cycle counter counter

	if (mExpChip) mExpChip->loadState(stream);

	mintSampleCtr = mintOSCtr = 1; // Reset these guys right away
}