//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#ifndef NESAPU
#define NESAPU

#include "Porting/basetypes.h"

#include "Cpu6502.h"
#include "NesCPU.h"

#include "SoundChips/BaseSoundChip.h"
#include "SoundChips/VRC6.h"


#define		COUNTMODE	1
#define		LOADMODE	0

#define		TRIMIN		0
#define		TRIMAX		15

// Enumerate all expansion chips
#define		VRC6CHIP	0x00

class NesCPU;

class NesAPU {

public:		
	///////////////////////////////
	// Class Procedures
	///////////////////////////////
	NesAPU();
	~NesAPU();
	void Init();

	unsigned char read(Uint16 addr);
	void write(Uint16 addr, unsigned char data);	
	void linkCPU(NesCPU *linkCPU);
	void reset();

	void exec(int intCycles);	
	void endFrame();
	
	void setMachineMode(int intNewMode);
	
	void addExpansionChip(int intExpChip);
	void writeExpansionChip(Uint16 addr, unsigned char data);

	void saveState(FILE *stream);
	void loadState(FILE *stream);

	///////////////////////////////
	// Class Data Members
	///////////////////////////////
	int mintFrameCounter;
	int mintClockDivider;
	int mintFrameCycleCtr;
	int mintFrameThirdCtr;
	int mintSampleCtr;
		
	char *mPlayBuffer;					// Pointer to current play sound buffer

	struct SquareChannel {
		bool	blnEnabled;					// Is channel enabled? (Writes to $4015)
		bool	blnLengthEnabled;			// Is length counter enabled?
		int		intLengthCtr;				// Length counter for square channel #1		
		
		bool	blnSweepEnabled;			// Is sweeping enabled?
		int		intSweepRate;				// Rate at which sweep refreshes ($4001)
		int		intSweepCtr;				// Sweep clock counter
		int		intShiftAmt;				// Amount to sweep by
		bool	blnSweepDir;				// Direction in which to sweep

		bool	blnDecayEnabled;			// Is decay enabled? ($4000)
		int		intDecayCtr;				// Decay counter for sq 1
		int		intDecayRate;				// Decay rate ($4000)
		int		intVolume;					// Set volume
		int		intCurVol;					// Current volume
		
		int		intDutyType;				// Type (ratio) of duty cycling
		int		intDutyCtr;					// Duty Cycle counter
		int		intWavelength;				// Current wavelength
		int		intCycleCtr;
		int		intCurSample;				// Current output
				
		bool	blnForceMute;				// Is channel forced to be muted? (from client)
	};

	struct TriangleChannel {
		bool	blnEnabled;					// Is channel enabled?
		bool	blnLengthEnabled;			// Is length counter enabled?
		int		intLengthCtr;				// Length counter
				
		int		intLinearCtr;				// Linear Counter
		int		intLinearLimit;				// Linear counter reset value
		int		intLinearMode;				// Mode linear counter is in (count or load)
		bool	blnLinearEnable;			// Is linear counter enabled?  (If not, length counter is)
		bool	blnChangeMode;

		int		intWavelength;				// Wavelength of triangle wave
		int		intCycleCtr;
		int		intCurSample;

		int		intStepCtr;					// Counter to which (0-31) phase of the triangle wave we are doing
		int		intStepInc;					// Increment to step by (1 or -1)		
		
		bool	blnForceMute;				// Is channel forced to be muted? (from client)
	};

	struct NoiseChannel {
		bool	blnEnabled;
		bool	blnLengthEnabled;
		int		intLengthCtr;
		int		intLengthLimit;
		
		bool	blnDecayEnabled;			// Is decay enabled? ($4000)
		int		intDecayCtr;				// Decay counter for sq 1
		int		intDecayRate;				// Decay rate
		int		intVolume;					// Set volume
		int		intCurVol;					// Current volume

		bool	blnDataType;				// Random number generation type		
		int		intCurRand;					// Current random number
		int		intCurSample;
		int		intCycleCtr;
		Uint16 intShiftReg;					// Shift register for random number gen

		Uint16 intWavelength;				// Wavelength

		bool	blnForceMute;
	};

	struct	DMCChannel {
		bool	blnEnabled;
		bool	blnForceMute;

		int		intDMAAddr;					// DMA address
		int		intDMAAddrLatch;			// Latch copy of DMA address

		int		intDMALength;				// Length of sample to play
		int		intDMALengthLatch;			// Latch copy of DMA length
		int		intDMARate;					// Rate of which to update sample
		int		intCycleCtr;				// Cycle counter for sample update
		int		intShiftCtr;				// Cycle counter for shift register
		int		intCurByte;					// Current byte fetched from DMA

		int		intDeltaCounter;			// Delta counter register
		int		intPlayMode;				// Play mode (3 types)
		int		intDAC;						// DAC output (7 bits)
		int		intCurSample;				// Current DMA sample

	};

	SquareChannel sq1, sq2;
	TriangleChannel tri;
	NoiseChannel nse;
	DMCChannel dmc;

	int intDMACycles[2][16];			// Lookup table for DMA frequency cycles
	int intNoiseCycles[16];				// Lookup table for noise wavelength
	int intLengthTable[32];				// Lookup table for length counter frames
	
	// The following variables are included to make the APU flexible for a client program
	bool mboolSoundEnabled;				// Is sound even enabled?
	int mintMasterVolume;				// Master volume level	
	int mintSampleRate;					// Output Sample rate (44100 default)
	int mintOSRate;						// Oversample factor (x4)
	int mintSampleSize;					// Sample size (in bits, 16 default, unused)
	int mintActualSamplesPerFrame;		// Actual number of samples generated per frame
	bool blnForceExpMute;				// Mute expansion chip?
	
private:
	// All-channel methods	
	void clockLengthCtr(int channel);	
	
	// Square channel methods
	void clockFreqSweep(SquareChannel &sq, int channel);	
	void clockVolumeDecay(SquareChannel &sq);
	void initSquareChannel(SquareChannel &sq);

	// Triangle channel methods
	void clockLinearCounter();
		
	// Noise Channel methods		
	void clockNoiseVolume();	

	// Methods for running each channel a given number of cycles
	void execSquare(SquareChannel &sq, int intCycles);
	void execTriangle(int intCycles);
	void execNoise(int intCycles);
	void execDMC(int intCycles);
		
	void clockFrameCounter();

	void initSoundBuffers();
		

	BaseSoundChip *mExpChip;			// Expansion sound chip object	
	
	bool	blnIRQEnabled;
	NesCPU *mCpu;
	
	int mintBufferSize;					// Size of sound buffer	
	int	mintCyclesPerSample;
	int mintCycleCtr;
	int mintOSCtr;
	int mintOutputCtr;	

	int intCPUSpeed;
	int intWantFPS;
	int intMachineMode;
	int intCyclesPerFrame;
	
	// Tables for mixing the 5 channels together (from blaarg's document)
	float square_table[32], tnd_table[204];

	float fltSample;
};

#endif
