//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// Nes CPU.cpp: implementation of the NesCPU class.
//////////////////////////////////////////////////////////////////////

#include "NesCPU.h"

//////////////////////////////////////////////////////////////////////
// Procedure: Init
//////////////////////////////////////////////////////////////////////
// Description: Sets the initial register values, initializes memory,
// and initializes some data members.
//////////////////////////////////////////////////////////////////////
void NesCPU::Init() {
	mPpu = NULL;
	mApu = NULL;
	mController1 = mController2 = NULL;

	// Provide initial values for registers
	reg.acc = 0;
	reg.x = 0;
	reg.y = 0;	
    reg.pc = 0x8000;		// Default code start location		
	reg.sp = 0xFF;			// Top  of stack
	reg.status = SET_UNUSED;

	mintInterruptLine = 0;	// Wait to raise RESET interrupt until a cartridge is loaded
			
	// Free memory if needed (in case of a system reset)
	if (memory != NULL)	delete memory;
	
	// Allocate and zero memory
	memory = new unsigned char[0xFFFF + 1];
	int intCtr = 0;
	while (intCtr <= 0xFFFF) { memory[intCtr] = 0; intCtr++; }
	
	mblnToggle = true;

	mlngCycleCounter = 0;
	mlngFrameCtr =  0;	

	MTimers.intCPUSpeed = 1789772;
	MTimers.intWantFPS = 60;
}

//////////////////////////////////////////////////////////////////////
// Procedure: setMachineMode
//////////////////////////////////////////////////////////////////////
// Set the NES machine mode, which is a series of timing values
//////////////////////////////////////////////////////////////////////
void NesCPU::setMachineMode(int intMachineMode) {	
	switch (intMachineMode) {
	case NTSCNESMODE:
		MTimers.intCyclesPerFrame = 29780;
		MTimers.intCyclesPerFrameNum = 2;
		MTimers.intCyclesPerFrameDen = 3;

		MTimers.intCyclesPerScanline = 113;
		MTimers.intCyclesPerScanlineNum = 2;
		MTimers.intCyclesPerScanlineDen = 3;

		MTimers.intCPUSpeed = 1789772;
		MTimers.intWantFPS = 60;

		if (mPpu) {
			mPpu->mintNumScanlines = 262;
			mPpu->mintNumVBlankScanlines = 20;
		}
		if (mApu) {
			mApu->setMachineMode(intMachineMode);
			mApu->Init();
		}
		break;

	case PALNESMODE:
		MTimers.intCyclesPerFrame = 33252;
		MTimers.intCyclesPerFrameNum = 7;
		MTimers.intCyclesPerFrameDen = 50;

		MTimers.intCyclesPerScanline = 106;
		MTimers.intCyclesPerScanlineNum = 9;
		MTimers.intCyclesPerScanlineDen = 16;
		
		MTimers.intCPUSpeed = 1662607;
		MTimers.intWantFPS = 50;
		if (mPpu) {
			mPpu->mintNumScanlines = 312;
			mPpu->mintNumVBlankScanlines = 70;
		}
		if (mApu) {
			mApu->setMachineMode(intMachineMode);
			mApu->Init();
		}
		break;
	}

	mintFrameCycleCtr = MTimers.intCyclesPerFrame;
	mintFrameFrac = MTimers.intCyclesPerFrameDen;
	mintScanlineCtr = MTimers.intCyclesPerScanline;	
	mintScanlineFrac = MTimers.intCyclesPerScanlineDen;
}


//////////////////////////////////////////////////////////////////////
// Procedure: RAM_Read
//////////////////////////////////////////////////////////////////////
// Description:  Reads a byte of memory from CPU RAM.  This is given
// a separate procedure because the CPU communicates with other
// devices via memory-mapped hardware, so there are many special
// cases here.
//////////////////////////////////////////////////////////////////////
unsigned char NesCPU::RAM_Read(Uint16 addr)
{	
	// Look at addr in $2000 blocks
	//switch ( (addr & 0xE000) >> 13 )
	switch ( addr >> 13 )
	{
		case 0x0: // $0000 to $1FFF is CPU memory (mirrored several times)		
			return memory[addr & 0x7FF];
			break;

		case 0x1:	// $2000 to $3FFF contains some PPU registers that are mirrored a bunch of times		
			switch ((addr & 0x7)+0x2000)
			{
				case 0x2002:	// PPU Status Register					
					{
						mblnToggle = true;				// reset toggle mechanism

						// Clear the VBlank bit after a read
						unsigned char bytRetVal = mPpu->regs.bytPPUStatus;
						mPpu->regs.bytPPUStatus &= 0x7F;
							
						// The status register is maintained in the PPU class
						return bytRetVal;		
					}
					break;
				
				case 0x2007:	// VRAM I/O Register
					return mPpu->VRAM->read( mPpu->regs.VRAMAddr, true);		
					break;
				
				default:	
					return 0;
					break;
			}
			break;

		case 0x2:	// $4000 to $5FFF contains some input registers and Expansion ROM		
			switch (addr & 0x3F40)
			{
			case 0x0000:
				switch (addr)
				{
				case 0x4016:	// Joypad #1 read
					return mController1->poll();
					break;
				
				case 0x4017:	// Joypad #2 read
					return mController2->poll();
					break;
				
				default:		// All other registers in this block are APU
					return mApu->read(addr);
					break;
				}
			default:	// Expansion ROM in $4020 to $5FFF				
				return mMapper->read(addr);
				break;
			}
			break;

		case 0x3:	// $6000 to $7FFF contains SRAM
			return memory[addr];
			break;
		
		default:	//... and the rest is just PRG-ROM
			//return memory[addr];			
			return mMapper->read(addr);
			break;					
	}
}



//////////////////////////////////////////////////////////////////////
// Procedure: RAM_Write
//////////////////////////////////////////////////////////////////////
// Description:  Writes a byte of data to CPU memory.  Like RAM_Read,
// this procedure has many special cases because the CPU communicates
// with the PPU, controllers, etc, by writing data to a certain memory
// location.
//////////////////////////////////////////////////////////////////////
void NesCPU::RAM_Write(Uint16 addr, unsigned char data)
{	
	
	// Look at addr in $2000 blocks
	switch ( addr >> 13 )
	{
	case 0x0: // $0000 to $1FFF is CPU memory (mirrored several times)
		memory[addr & 0x7FF] = data;
		break;

	case 0x1: // $2000 to $3FFF contains several PPU registers
		switch ((addr & 0x7)+0x2000)
		{
		case 0x2000:	// PPU control register #1 (write only)			
			mPpu->regs.bytControlReg1 = data;

			// Copy the 2 name table bits to the temp VRAM address register
			mPpu->regs.tempVRAMAddr = (mPpu->regs.tempVRAMAddr & 0xF3FF) | (data & 0x3) << 10;		
			
			// Set PPU increment
			if (data & 0x04) {
				mPpu->VRAM->mintIncrement = 32;
			} else {
				mPpu->VRAM->mintIncrement = 1;
			}

			// If they enable NMI IRQs while the VBlank flag is set, trigger the NMI
			/*if ((data & 0x80) && (mPpu->regs.bytPPUStatus & 0x80)) {
				mintInterruptLine |= NMI;
			}*/

			break;		

		case 0x2001:	// PPU control register #2 (write only)		
			mPpu->regs.bytControlReg2 = data;
			break;
	
		case 0x2002:	// PPU status is read only
			break;
		
		case 0x2003:
			mPpu->regs.bytSPRRAMAddr = data;
			break;
		
		case 0x2004:
			mPpu->SPRRAM[ mPpu->regs.bytSPRRAMAddr ] = data;
			break;
		
		case 0x2005:	// Temp VRAM address register, used for scrolling			
			if (mblnToggle) {
				// First write - The upper 5 bits of the data are copied to bits 0-5 of the temp vram addr
				mPpu->regs.tempVRAMAddr = (mPpu->regs.tempVRAMAddr & 0xFFE0) | ((data & 0xF8) >> 3);
				mPpu->draw.bytXPixelOffset = data & 0x7;	// Lower 3 bits of data are x pixel offset			
			} else {
				// Second write - The lower 3 bits of the data are copied to bits 12-14 and the upper 5 bits to bits 5-9
				mPpu->regs.tempVRAMAddr = (mPpu->regs.tempVRAMAddr & 0xFC1F) | ((data & 0xF8) << 2);
				mPpu->regs.tempVRAMAddr = (mPpu->regs.tempVRAMAddr & 0x8FFF) | ((data & 0x7) << 12);			
			}			 
			mblnToggle = !mblnToggle;
			break;
		
		case 0x2006:	// VRAM internal address register			
			if (mblnToggle) {
				// First write - the lower 6 bits of data are copied to temp vram register bits 8-13
				// Note: bits 14 and 15 are set to zero as well
				//mPpu->regs.tempVRAMAddr = (mPpu->regs.tempVRAMAddr & 0x00FF) | ((data & 0x3F) << 8);
				mPpu->regs.tempVRAMAddr = (mPpu->regs.tempVRAMAddr & 0xC0FF) | ((data & 0x3F) << 8);
				mPpu->regs.tempVRAMAddr &= 0x3FFF;
			} else {
				// Second write, the entire byte of data is copied to lower byte of temp address, 
				// then the temp address is copied to the "real" VRAM address
				mPpu->regs.tempVRAMAddr = (mPpu->regs.tempVRAMAddr & 0xFF00) | data;
				mPpu->regs.VRAMAddr = mPpu->regs.tempVRAMAddr;			
			}						
			mblnToggle = !mblnToggle;
			break;
		
		case 0x2007:	// VRAM I/O Register
			// Write byte to the current VRAM address
			mPpu->VRAM->write(mPpu->regs.VRAMAddr, data, true);		
			break;
		}
		break;

	case 0x2:	// $4000 to $5FFF contains several registers
		switch (addr & 0x3F40)	// isolate down to $4000 to $4020
		{
			case 0x0000:	// $4000 to $4020 are registers
				switch (addr)
				{
				case 0x4014:	// Sprite DMA Register
					doDMATransfer (data);
					break;
				
				case 0x4016:	// Joypad #1 register (allegedly?)
					if (data & 0x01)
					{
						mController1->clearStrobe();
						mController2->clearStrobe();
					} else {			
						mController1->resetStrobe();
						mController2->resetStrobe();
					}
					break;
				
				/*case 0x4017:	// Joypad #2 register (allegedly?)
					break;*/
				
				default:	// All other registers in this block are APU regs
					mApu->write(addr, data);
					break;
				}
				break;

			default:	// $4020 to $5FFF is expansion ROM; send to mapper
				//memory[addr] = data;
				mMapper->write(addr, data);
				break;
		}
		break;
			
	case 0x3:	// $6000 to $7FFF is SRAM				
		memory[addr] = data;			// Some mappers (e.g. 34) use this address space
		mMapper->write(addr, data);		// so we do the write then send it to the mapper
		break;
	
	default:	// PRG-ROM: a write to PRG-ROM triggers the mapper interface
		mMapper->write(addr, data);
		break;
	}
	
}



//////////////////////////////////////////////////////////////////////
// Procedure: calcFrame
//////////////////////////////////////////////////////////////////////
// Description:  Calculates one frame of NES action, ideally this would
// be done 60 times per second.  Think of it like this:  the client calls
// calcFrame, and the Nes object's PPU->display bitmap is drawn in with
// that frame's graphics, and the Nes object's APU->mPlayBuffer is filled
// with 1/60 of a second of sound.
//////////////////////////////////////////////////////////////////////
void NesCPU::calcFrame() {			  
	
	bool blnVINT = false;
	int Spr0RaiseCycle = 0, OtherIRQRaiseCycle = 0;
	int vCycles = 4;
	while (mintFrameCycleCtr > 0) {
		// Draw scanline		
		mPpu->timers.intSpr0HitPix = -1;		
		mPpu->calcScanLine();

		// Raise VINT status bit 
		if (mPpu->mintCurScanLine == 1) {
			mPpu->regs.bytPPUStatus |= 0x80;
		}
		
		// Trigger the NMI flag
		if ((mPpu->regs.bytPPUStatus & 0x80) && (blnVINT == false) && (mPpu->regs.bytControlReg1 & 0x80)) {
		//if ((mPpu->regs.bytControlReg1 & 0x80) && (mPpu->mintCurScanLine == 21) && !(blnVINT)) {
			// Raise a NMI interrupt
			//mintInterruptLine |= NMI;
			blnVINT = true;			
		}

		// Check for other IRQs from PPU
		if (mPpu->timers.blnOtherIRQ) {
			OtherIRQRaiseCycle = mintFrameCycleCtr - ( (mPpu->timers.intOtherPix + (16 - mPpu->draw.bytXPixelOffset)) / 3);
			mPpu->timers.blnOtherIRQ = false;
		}
	
		// Check for spr#0 hits and set off counter if so
		if (mPpu->timers.intSpr0HitPix != -1) {			
			Spr0RaiseCycle = mintFrameCycleCtr - ( (mPpu->timers.intSpr0HitPix + (16 - mPpu->draw.bytXPixelOffset)) / 3); 			
		}


		// Execute CPU time for given scanline
		while (mintScanlineCtr > 0) {
			executeInstruction();

			mintScanlineCtr -= mlngCycleCounter;			

			// Clock frame cycle counter
			mintFrameCycleCtr -= mlngCycleCounter;

			if (blnVINT && (vCycles > 0)) {
				vCycles -= mlngCycleCounter;
				if (vCycles <= 0) {
					mintInterruptLine |= NMI;					
				}
			}

			// See if we should raise the sprite #0 flag
			if (Spr0RaiseCycle) {
				if (mintFrameCycleCtr <= Spr0RaiseCycle) {
					// Raise flag					
					mPpu->regs.bytPPUStatus |= 0x40;
					Spr0RaiseCycle = 0;
				}
			}

			// See if we should raise any other IRQs from PPU
			if (OtherIRQRaiseCycle) {
				if (mintFrameCycleCtr < OtherIRQRaiseCycle) {
					mintInterruptLine |= IRQ;
					OtherIRQRaiseCycle = 0;					
				}
			}

			// Run APU
			mApu->exec( mlngCycleCounter);
			mlngCycleCounter = 0;
		}
				
		// Clock scanline counter
		mintScanlineFrac--;		
		if (!mintScanlineFrac) mintScanlineFrac = MTimers.intCyclesPerScanlineDen;
		if (mintScanlineFrac > MTimers.intCyclesPerScanlineNum) {
			mintScanlineCtr = MTimers.intCyclesPerScanline + mintScanlineCtr;
		} else {
			mintScanlineCtr = MTimers.intCyclesPerScanline + 1 + mintScanlineCtr;
		}
	}

	// Reset frame counter
	mintFrameFrac--;
	if (!mintFrameFrac) mintFrameFrac = MTimers.intCyclesPerFrameDen;
	if (mintFrameFrac > MTimers.intCyclesPerFrameNum) {
		mintFrameCycleCtr = MTimers.intCyclesPerFrame + mintFrameCycleCtr;
	} else {
		mintFrameCycleCtr = MTimers.intCyclesPerFrame + 1 + mintFrameCycleCtr;
	}
	
	// Run APU until end...
	mApu->endFrame();
	
	mlngFrameCtr++;
}



//////////////////////////////////////////////////////////////////////
// Procedure: doDMATransfer
//////////////////////////////////////////////////////////////////////
// Description: Transfers a page (256 bytes) of CPU memory to the special
// SPRRAM page.  (SPRRAM is a feature of the NesPPU class)
//////////////////////////////////////////////////////////////////////
void NesCPU::doDMATransfer(unsigned char page)
{
	// Shift the page value over 8 bits
	int intAddr = page * 0x100;
		
	// Loop through and copy memory
	int ctr = 0x00;
	while (ctr <= 0xFF)
	{
		mPpu->SPRRAM[ctr] = RAM_Read(intAddr + ctr);
		ctr++;
	}
	mlngCycleCounter += 512;
}



//////////////////////////////////////////////////////////////////////
// Procedure: handleInterruptRequest
//////////////////////////////////////////////////////////////////////
// Description: Handles interrupts for the NES.
//////////////////////////////////////////////////////////////////////
void NesCPU::handleInterruptRequest()
{		
	// Standard interrupt opening: push PC and status onto stack, 
	memory[reg.sp + 0x100] = ((reg.pc) & 0xFF00) >> 8;	    // PCH to reg.sp
	memory[reg.sp - 1 + 0x100] = (reg.pc) & 0xFF;	  	    // PCL to reg.sp-1			
	memory[reg.sp - 2 + 0x100] = reg.status;				// Status to reg.sp-2	
	reg.sp = reg.sp - 3;									// Adjust stack pointer
	mlngCycleCounter += 7;									// 7 clock cycles to start int
	
	if (mintInterruptLine & RESET) {
		// Set RESET vector
		reg.pc = memory[0xFFFC] + (memory[0xFFFD] << 8);
		mApu->reset();
		mintInterruptLine = 0;
	} else if (mintInterruptLine & NMI)	{
		// Set NMI vector
		reg.pc = memory[0xFFFA] + (memory[0xFFFB] << 8);
		mintInterruptLine = 0;
	} else if (mintInterruptLine & IRQ)	{
		// Set IRQ/BRK vector if interrupt flag is down
		if (!(reg.status & SET_INTERRUPT)) {
			reg.pc = memory[0xFFFE] + (memory[0xFFFF] << 8);
			mintInterruptLine = 0;
			reg.status |= SET_INTERRUPT; // raise interrupt flag
		}
	}/* else if (mintInterruptLine & BRK) {
		if (!(reg.status & SET_INTERRUPT)) {
			// Set IRQ/BRK vector IF the interrupts are enabled		
			reg.pc = memory[0xFFFE] + (memory[0xFFFF] << 8);
			mintInterruptLine = 0;
			reg.status |= SET_INTERRUPT;
		}
	}*/
}



//////////////////////////////////////////////////////////////////////
// Procedure: linkCPU
//////////////////////////////////////////////////////////////////////
// Description: Links the NesCPU object with its "sibling" objects, such
// as the NesPPU, controllers, memory mappers, etc.  This is a bit of a
// contrived way to do things, but think of it like this:  The Nes class
// represents the entire NES.  It is the parent.  It's children (datamembers)
// are all of the subcomponents, like the CPU, PPU, etc.  These children 
// all must work together, and the way they do it is by sharing their 
// pointers with one another.
//////////////////////////////////////////////////////////////////////
void NesCPU::linkCPU(NesPPU *ppu, NesAPU *apu, NesInput *controller1, NesInput *controller2, NesBaseMapper *mapper)
{
	this->mPpu = ppu;
	this->mApu = apu;
	this->mController1 = controller1;
	this->mController2 = controller2;
	this->mMapper = mapper;
}



//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the NES CPU to the file stream
//-------------------------------------------------------------------
void NesCPU::saveState(FILE *stream) {
	
	fwrite(&memory[0x6000], 0x2000, 1, stream); // SRAM
	fwrite(&memory[0x0000], 0x800,  1, stream); // RAM
	fwrite(&reg, sizeof(regs), 1, stream);		// Registers
	fwrite(&mintFrameCycleCtr, sizeof(mintFrameCycleCtr), 1, stream);		// Cycle counter
	fwrite(&mintFrameFrac, sizeof(mintFrameFrac), 1, stream);		// Cycle counter
	fwrite(&mintScanlineCtr, sizeof(mintScanlineCtr), 1, stream);			// Cycle counter
	fwrite(&mintScanlineFrac, sizeof(mintScanlineFrac), 1, stream);		// Cycle counter
	fwrite(&mblnToggle, sizeof(mblnToggle), 1, stream);						// Misc
	fwrite(&mlngCycleCounter, sizeof(mlngCycleCounter), 1, stream);			// Misc
	fwrite(&mintInterruptLine, sizeof(mintInterruptLine), 1, stream);		// Misc
}




//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the NES CPU from the file stream
//-------------------------------------------------------------------
void NesCPU::loadState(FILE *stream) {	
	fread(&memory[0x6000], 0x2000, 1, stream); // SRAM
	fread(&memory[0x0000], 0x800,  1, stream); // RAM
	fread(&reg, sizeof(regs), 1, stream);	   // Registers	
	fread(&mintFrameCycleCtr, sizeof(mintFrameCycleCtr), 1, stream);		// Cycle counter
	fread(&mintFrameFrac, sizeof(mintFrameFrac), 1, stream);		// Cycle counter
	fread(&mintScanlineCtr, sizeof(mintScanlineCtr), 1, stream);			// Cycle counter
	fread(&mintScanlineFrac, sizeof(mintScanlineFrac), 1, stream);		// Cycle counter
	fread(&mblnToggle, sizeof(mblnToggle), 1, stream);						// Misc
	fread(&mlngCycleCounter, sizeof(mlngCycleCounter), 1, stream);			// Misc
	fread(&mintInterruptLine, sizeof(mintInterruptLine), 1, stream);		// Misc
}


// Provide this so sub-classes can override this method
void NesCPU::executeInstruction() {
	cpu6502::executeInstruction();
}