//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// Nes CPU.h: interface for the NesCPU class.
//////////////////////////////////////////////////////////////////////
// Class Description:  The NesCPU class is derived from the cpu6502 class.
// It provides a NES-specific initialization procedure, interrupt handler,
// and CPU memory access procedures.  It also provides a procedure that 
// calculates a frame of NES computation, a frame being what a NES calculates
// every 1/60 of a second.
//
// This class is heavily linked with other NES classes.  Specifically, 
// access to the NES PPU and controller classes are required in order
// for the CPU to properly do its job.
//////////////////////////////////////////////////////////////////////

#ifndef NES_CPU
#define NES_CPU

#include "Porting/basetypes.h"

#include "Cpu6502.h"
#include "NesPPU.h"
#include "NesInput.h"

#define	NTSCNESMODE		0
#define	PALNESMODE		1


// Because all of these classes work together, sometimes we need to use some trickery
// to get all of the definitions in
class NesBaseMapper;
class NesAPU;



class NesCPU : public cpu6502 {
public:
	///////////////////////////////
	// Class Procedures
	///////////////////////////////
	NesCPU() {};
	virtual ~NesCPU() {	};
	
	void Init();														// override method	
	virtual unsigned char RAM_Read(Uint16 addr);				  		// override method
	virtual void RAM_Write(Uint16 addr, unsigned char data);       // override method
	void handleInterruptRequest();									// override method

	virtual void executeInstruction();

	virtual void calcFrame();
	void doDMATransfer(unsigned char page);
	void linkCPU(NesPPU *realNesPPU, NesAPU *realNesApu, NesInput *controller1, NesInput *controller2, NesBaseMapper *mapper);

	void setMachineMode(int intMachineMode);
	
	void saveState(FILE *stream);
	void loadState(FILE *stream);
	

	///////////////////////////////
	// Class Data Members	
	///////////////////////////////	
	long mlngFrameCtr;		
	bool mblnToggle;

	// These are links to all of the components of the NES
	NesPPU  *mPpu;	
	NesAPU	*mApu;
	NesInput *mController1;
	NesInput *mController2;

	// This structure holds the timing values for the machine.  The values
	// are different for NTSC vs. PAL NES'
	struct MachineTimers {
		int intCyclesPerFrame;			// CPU cycles per frame integer
		int intCyclesPerFrameNum;		//	+ numerator
		int intCyclesPerFrameDen;		//  + denominator

		int intCyclesPerScanline;		// CPU cycles per scanline
		int intCyclesPerScanlineNum;	//  + numerator
		int intCyclesPerScanlineDen;	//  + denominator

		int intCPUSpeed;				// CPU speed
		int intWantFPS;					// Desired FPS
	};
	MachineTimers MTimers;	


protected:
	NesBaseMapper *mMapper;
	int mintFrameCycleCtr;
	int mintFrameFrac;
	int mintScanlineCtr;
	int mintScanlineFrac;


private:
	int mintCurMachineType;

};

// Need to include this after the clas is defined
#include "NesBaseMapper.h"
#include "NesAPU.h"
#endif
